
#include <USBKit.h>
#include <stdio.h>
#include <string.h>

/* Vendor Control Ops for the Cypress Thermo Board */
#define THERMO_OP_PING      0x00
#define THERMO_OP_READ_RAM  0x02
#define THERMO_OP_WRITE_RAM 0x03

/* Interesting Memory Locations on the Cypress Thermo Board */
#define THERMO_SENSOR0      0x78
#define THERMO_SENSOR1      0x79
#define THERMO_BUTTON       0x7A
#define THERMO_LED_LEVEL    0x2c
#define THERMO_LED_CHANGED  0x2b

class ThermoWatcher
{
public:
	ThermoWatcher(USBDevice *therm){
		d = therm;
		fprintf(stderr,"ThermoWatcher#%d: Hello, USB!\n",counter);
		num = counter++;
		if(therm->SetConfiguration(therm->ConfigurationAt(0))){
			running = 0;
			t = -1;
		} else {
			running = 1;
			t = spawn_thread(ThermoWatcher::ThreadEntry,"ThermoWatcher",
						 B_NORMAL_PRIORITY,this);
			resume_thread(t);
		}
	}
	~ThermoWatcher(){
		if(t >= 0){
			status_t res;
			running = 0;
			wait_for_thread(t,&res);
			t = -1;
		}
		printf("ThermoWatcher#%d: Goodbye Cruel World.\n",num);
	}
	USBDevice *GetDevice(void){
		return d;
	}
	void SetLED(uint level){
		d->ControlTransfer(USB_REQTYPE_VENDOR|USB_REQTYPE_ENDPOINT_IN,
						   THERMO_OP_WRITE_RAM,THERMO_LED_LEVEL,level,0,0);
		d->ControlTransfer(USB_REQTYPE_VENDOR|USB_REQTYPE_ENDPOINT_IN,
						   THERMO_OP_WRITE_RAM,THERMO_LED_CHANGED,0x01,0,0);
	}
	int ReadButton(void){
		uchar data[8];
		d->ControlTransfer(USB_REQTYPE_VENDOR|USB_REQTYPE_ENDPOINT_IN,
						   THERMO_OP_READ_RAM,THERMO_BUTTON,0x00,8,data);
		return data[1];
	}
	int ReadTemp(void){
		int t;
		uchar data[8];
		if(d->ControlTransfer(USB_REQTYPE_VENDOR|USB_REQTYPE_ENDPOINT_IN,
							  THERMO_OP_READ_RAM,THERMO_SENSOR1,0x00,8,data)){
			return -1;
		}
							  
		t = data[1];
		if(d->ControlTransfer(USB_REQTYPE_VENDOR|USB_REQTYPE_ENDPOINT_IN,
							  THERMO_OP_READ_RAM,THERMO_SENSOR0,0x00,8,data)){
			return -1;
		}					  
		return (t << 8) | data[1];
	}
	status_t Thread(void){
		uint br = 0;
		while(running){
			// Blink the LED
			if(br) { br = 0; } else { br = 255; }  SetLED(br);
			
			// Display Button State and Temperature
			printf("Thermo#%d: Button %s / Temp %d\n",
				   num,ReadButton() ? "On " : "Off",ReadTemp());
			
			// Wait a second
			snooze(1000000);
		}
		return 0;
	}
	static status_t ThreadEntry(void *data){
		return ((ThermoWatcher *) data)->Thread();
	}
	
private:
	static int counter;
	int num;
	int running;
	thread_id t;
	USBDevice *d;
};

int ThermoWatcher::counter = 0;



class Roster : public USBRoster
{
	BList devs;
		
public:
	virtual status_t DeviceAdded(USBDevice *dev){
		if((dev->VendorID() == 0x04b4) && (dev->ProductID() == 0x0002)){
			// It's a Cypress Thermometer Demo Board, cool!
			devs.AddItem(new ThermoWatcher(dev));
			return B_OK;
		} else {
			// Some other critter that we don't care about.
			return B_ERROR;
		}
	}
	virtual void DeviceRemoved(USBDevice *dev){
		int i;
		ThermoWatcher *t;
		for(i=0;t=(ThermoWatcher *)devs.ItemAt(i);i++){
			if(t->GetDevice() == dev){
				devs.RemoveItem(i);
				delete t;
			}
		}
	}	
};

int main(int argc, char *argv[])
{
	Roster roster;
	roster.Start();
	while(1) snooze(1000000); // wait around and let the roster do its thing
	return 0;
}

